@@ -1,11 +1,9 @@
 import json
-from pydantic import BaseModel
-from typing import Any, Dict, Optional, List, Union
+from typing import Any, Dict, Optional, List, Callable
 from xpander_sdk import (
     ToolCall,
     ToolCallType,
     ToolCallResult,
-    ILocalTool,
     LLMProvider,
 )
 from .base import SDKAdapter
@@ -121,7 +119,7 @@ class ChainlitAdapter(SDKAdapter):
                     tool_calls[tc.index] = ToolCall(
                         name=tc.function.name,
                         tool_call_id=tc.id,
-                        type=ToolCallType.XPANDER,
+                        type=ToolCallType.XPANDER if not tc.function.name.startswith("xpLocal") else ToolCallType.LOCAL,
                         payload="",
                     )
                 else:
@@ -129,7 +127,8 @@ class ChainlitAdapter(SDKAdapter):
 
         return tool_calls
 
-    def process_tool_calls(self, tool_calls: Dict[int, ToolCall]):
+    @cl.step(type="tool")
+    def process_tool_calls(self, tool_calls: Dict[int, ToolCall], local_tools: Dict[str, Callable] = None):
         """
         Process tool calls by formatting their payloads and executing them.
 
@@ -158,8 +157,39 @@ class ChainlitAdapter(SDKAdapter):
                 }
             ]
         )
-
-        self.agent.run_tools(tool_calls=tool_calls_list)
+        
+        for tool_call in tool_calls.values():
+            current_step = cl.context.current_step
+            current_step.name = tool_call.name
+            current_step.input = tool_call.payload
+            
+            # handle local tools
+            if tool_call.type == ToolCallType.LOCAL:
+                tool_name = tool_call.name.replace("xpLocal_","")
+                if not local_tools:
+                    raise Exception(f"local_tools not initialized")
+                
+                tool_fn = local_tools[tool_name] if tool_name in local_tools else None
+                if not tool_fn:
+                    raise Exception(f"Tool {tool_name} implementation not found!")
+                
+                # run the local tool
+                tool_call_result = ToolCallResult(function_name=tool_call.name,tool_call_id=tool_call.tool_call_id,payload=tool_call.payload,result="",is_success=False,is_error=False)
+                try:
+                    fn_result = tool_fn(**tool_call.payload)
+                    tool_call_result.result = fn_result
+                    tool_call_result.is_success = True
+                    
+                except Exception as e:
+                    tool_call_result.result = str(e)
+                    tool_call_result.is_error = True
+                
+                # report result to the memory
+                self.agent.add_tool_call_results(tool_call_results=[tool_call_result])
+            else:
+                tool_call_result = self.agent.run_tool(tool=tool_call)
+            current_step.output = tool_call_result.result
+            current_step.language = "json"
 
         # Reset tool calls
         tool_calls.clear()
